package timertests;

/**
 * The class to organize and run the timing tests on a variety
 * of sorting algorithms, <CODE>Comaparator</CODE>s used, and 
 * the datasets to be sorted.
 * 
 * @(#)TimerTests  1.0  22 March 2006
 * @author Viera K. Proulx
 */

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Date;

import stresstests.datasets.*;
import stresstests.algorithms.*;
import stresstests.interfaces.*;

class TimerTests{
  
  /** compare cities by the name of the city */
  protected Comparator<City> byName = new ComparatorByName();
  
  /** compare cities by the zip code */
  protected Comparator<City> byZip = new ComparatorByZip();
  
  /** compare cities by their latitude */
  protected Comparator<City> byLatitude = new ComparatorByLatitude();

  /** create a list of all <CODE>Comparator</CODE>s */
  protected ArrayList<Comparator<City>> comparators = 
    new ArrayList<Comparator<City>>();
  
  /** create a corresponding list of all <CODE>Comparator</CODE> names */
  protected ArrayList<String> compNames = new ArrayList<String>();

  /** an <CODE>ArrayList</CODE> of algorithms to run */
  protected ArrayList<ASortAlgo<City>> algorithms = 
    new ArrayList<ASortAlgo<City>>();

  /** The collection of data sets used in the tests */
  protected TestData testdata = new TestData();
  
  /** the starting time for the sorting test */
  protected long startTime;  
  
  /** the ending time for the sorting test */
  protected long endTime;           
  
  /** 
   * The constructor only creates the three known <CODE>Comparator</CODE>s
   */
  protected TimerTests(){ 
    makeComparators();
  }
  
  /**
   * Construct the <CODE>ArrayList</CODE> of all available
   * <CODE>Comparator</CODE>s
   */
  protected void makeComparators(){
    comparators.add(byName);
    compNames.add("By Name");
    comparators.add(byZip);
    compNames.add("By Zip");
    comparators.add(byLatitude);
    compNames.add("By Latitude");
  }
  
  /**
   * Add to the <CODE>ArrayList</CODE> of <CODE>ASortAlgo</CODE>s
   * each available algorithm with the given <CODE>Comparator</CODE>
   * 
   * @param comp the <CODE>Comparator</CODE> used by these algorithms
   * @effect add an instance of each algorithm with the given 
   * <CODE>Comparator</CODE> to the <CODE>algorithms</CODE> 
   * <CODE>ArrayList</CODE>
   */
  protected void makeAlgos(Comparator<City> comp){
    // COMMENT OUT THE ALGORITHMS YOU DO NOT WANT TO RUN 
    algorithms.add(new ArrSortSelection<City>(comp));
    algorithms.add(new ArrSortInsertion<City>(comp));
    algorithms.add(new AListSortInsertion<City>(comp));
    algorithms.add(new AListSortQuickSort<City>(comp));
    algorithms.add(new ArrSortQuick<City>(comp));
  } 
  
  /**
   * Add to the <CODE>algorithms</CODE> each of the known algorithm 
   * with each of the given <CODE>Comparator</CODE>s
   * @param comparators
   */
  protected void allAlgos(ArrayList<Comparator<City>> comparators){
    for (Comparator<City> c : comparators){
      makeAlgos(c);
    }
  }
  
  /**
   * Produce a list of indices selecting the datasets
   * to use in the sorting tests
   * 
   * @return a list of indices indicating which datasets should be sorted
   */
  // SELECT THE DATASETS TO USE
  // COMMENT OUT THE INDICES FOR DATA SETS NOT TO BE USED 
  // IN THE CURRENT TEST RUN
  protected ArrayList<Integer> selectDataSets(){
    ArrayList<Integer> datasets = new ArrayList<Integer>();
    datasets.add(0);  // size 1000 sequential
    datasets.add(1);  // size 1000 random
    datasets.add(2);  // size 2000 sequential
    datasets.add(3);  // size 2000 random
    //datasets.add(4);  // size 4000 sequential
    //datasets.add(5);  // size 4000 random
    //datasets.add(6);  // size 8000 sequential
    //datasets.add(7);  // size 8000 random
    //datasets.add(8);  // size 16000 sequential
    //datasets.add(9);  // size 16000 random
    //datasets.add(10);  // size 29470 sequential
    
    return datasets;   
  }
    
  /**
   * Run the tests of all selected algorithms with all selected
   * comparators on all selected datasets and save the results
   * 
   * @return an <CODE>ArrayList</CODE> of <CODE>Result</CODE> data
   */
  protected ArrayList<Result> allTests(){
    // construct an ArrayList to hold the results
    ArrayList<Result> results = new ArrayList<Result>();
    
    // construct the list of all algorithms to try
    // with all chosen Comparators
    allAlgos(comparators);
    
    System.out.println("Number of comparators: " + comparators.size());
    System.out.println("Number of algorithms:  " + algorithms.size());  
    
    // select which datasets will be sorted
    ArrayList<Integer> selectedDatasets = selectDataSets();

    // for each algorithm in the algorithms ArrayList
    for (ASortAlgo<City> algo : algorithms){
      
      // and for the selected datasets
      for (int dindex : selectedDatasets)
        
        // run one timer test and add the result to the results ArrayList
        results.add(
            runATest(algo, testdata.datasets.get(dindex)));
    }
    
    // print the test results for all tests
    System.out.println("" + results.size() + " tests completed:");
    
    for (Result r : results)
      System.out.println(r.toString());
    return results;
  }
  
  /**
   * Run one test of a given sorting algorithm with the given comparator
   * and the desired size of data, random or sequentially selected
   * 
   * @param algo the <CODE>ASortAlgo</CODE> algorithm with the chosen 
   * <CODE>Comparator</CODE>
   * @param dataset the data to be sorted
   * 
   * @return an instance of the <CODE>Result</CODE> reporting all
   * relevant data
   */
  protected Result runATest(ASortAlgo<City> algo, DataSet<City> dataset) {
    
    // variable to record the sorted dataset
    Traversal<City> sorted;
    
    // initialize the data for the algorithm
    algo.initData(dataset.getTraversal());
    
    // run the timed test and return the measured time
    this.startTime = (new Date()).getTime();
    sorted = algo.sort();
    this.endTime = (new Date()).getTime();
    
    if (isSorted(sorted, algo.comp)){
    return new Result(dataset.size, 
                      dataset.type, 
                      algo.algoName, 
                      algo.comp.getClass().getName(),
                      this.endTime - this.startTime);
    }
    else{
      System.out.println(
          (new Result(dataset.size, 
                      dataset.type, 
                      algo.algoName, 
                      algo.comp.getClass().getName(),
                      this.endTime - this.startTime)).toString());
      throw new UnsortedException("The data not sorted");
    }
  }
  
  /**
   * A sample test for this class
   */
  protected void sampleTest(){
    
    // example of running the test for one choice of
    // algorithm, data to sort, and comparator
    long runtime = (this.runATest(
                         new ArrSortSelection(this.byName), 
                         testdata.datasets.get(0))).runtime;
    
    // we should print more info here - size of the data set,
    // is it random or sequential
    System.out.println("Selection sort ArrayList time: " + 
        runtime + "\n");
    
    System.out.println((this.runATest(
        new ArrSortSelection(this.byName), 
        testdata.datasets.get(2))).toString());
  }
  
  /**
   * Is the data generated by the given traversal sorted by comparator?
   * 
   * @param tr the given <CODE>Traversal</CODE> that generates the data
   * to be sorted
   * @param comp the <CODE>Comparator</CODE> that determines the ordering
   * 
   * @return true if the data is sorted
   */
  protected boolean isSorted(Traversal<City> tr, Comparator<City> comp){
    if (!tr.isEmpty()){
      return isSortedAcc(tr.getRest(), comp, tr.getFirst());
    }
    else
      return true;
  }

  /**
   * Is the data generated by the given traversal sorted by comparator?
   * and greater than or equal to the previous largest value?
   * 
   * @param tr the given <CODE>Traversal</CODE> that generates the data
   * to be sorted
   * @param comp the <CODE>Comparator</CODE> that determines the ordering
   * @param acc the last value seen traversing over the data
   * 
   * @return true if the data is sorted
   */
  protected boolean isSortedAcc(Traversal<City> tr, 
                             Comparator<City> comp,
                             City acc){
    if (!tr.isEmpty()){
      if (comp.compare(acc, tr.getFirst()) > 0)
        System.out.println("sort test failed: " + acc + " first = " + tr.getFirst());
      return (comp.compare(acc, tr.getFirst()) <= 0)
          && isSortedAcc(tr.getRest(), comp, tr.getFirst());
    }
    else
      return true;
  }   
  
  /**
   * Self Test
   */
  public static void main(String[] argv){
    TimerTests tt = new TimerTests();
    // tt.sampleTest();
    
    
    City northville = new City(48167, "Northville", "MI", 83.479355, 42.426245);
    City plymouth = new City(48170, "Plymouth", "MI", 83.479946, 42.36882);
    City gibraltar = new City(48173, "Gibraltar", "MI", 83.216196, 42.07918);
    City romulus = new City(48174, "Romulus", "MI", 83.358288, 42.220304);
    City saline = new City(48176, "Saline", "MI", 83.784936, 42.169844);
    City duluth1 = new City(55806, "Duluth", "MN", 92.127871, 46.771457);
    City duluth2 = new City(55807, "Duluth", "MN", 92.169821, 46.740783);
     
    AList<City> empty = new MTList<City>();
    
    AList<City> longunsorted = new ConsList<City>(northville,
        new ConsList<City>(plymouth,
            new ConsList<City>(romulus,
                new ConsList<City>(duluth1,
                    new ConsList<City>(gibraltar,
                        new ConsList<City>(saline,
                            new ConsList<City>(duluth2, empty)))))));

   
    AList<City> longsorted = new ConsList<City>(duluth1,
        new ConsList<City>(duluth2,
            new ConsList<City>(gibraltar,
                new ConsList<City>(northville,
                    new ConsList<City>(plymouth,
                        new ConsList<City>(romulus,
                            new ConsList<City>(saline, empty)))))));
    System.out.println("sorted test - success:" + 
        tt.isSorted(new TraversalList<City>(longsorted),
                    new ComparatorByName()));

    System.out.println("sorted test - failure:" + 
        tt.isSorted(new TraversalList<City>(longunsorted),
                    new ComparatorByName()));
    
    ArrayList<Result> results = tt.allTests();
    System.out.println("Completed " + results.size() + " tests.");
 
  }
}